# -*- coding: utf-8 -*-


import logging
import traceback

from .singleton import Singleton as dpSingleton


class Logger(dpSingleton):
    def __init__(self, engine=None):
        self.engine = engine

    def strip(self, msg, strip=False):
        return msg.strip() if strip else msg

    def exception(self, exception, *args, **kwargs):
        traceback.print_exc()

        logging.exception(self.strip(str(exception), True), *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        logging.error(self.strip(str(msg), True), *args, **kwargs)

    def info(self, msg, *args, **kwargs):
        logging.info(self.strip(msg), *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        logging.warning(self.strip(msg), *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        logging.debug(self.strip(msg), *args, **kwargs)



# -*- coding: utf-8 -*-


import tornado.web

from dp_tornado.engine.engine import Engine as dpEngine


class Pagination(tornado.web.UIModule, dpEngine):
    region_tag = 'div'
    region_class = 'pagination'

    first = ''
    first_class = ''

    prev_block = ''
    prev_block_class = ''

    next_block = ''
    next_block_class = ''

    prev = '<span></span> Prev'
    prev_class = 'direction prev'

    next = 'Next <span></span>'
    next_class = 'direction next'

    current_tag = 'strong'
    current_class = ''

    last = ''
    last_class = ''

    link_tag = 'a'
    link_class = ''
    link_prefix = ''
    link_suffix = ''
    link_function = ''

    space = ''

    def render(self, total_count, page, rpp, **options):
        rpb = options['rpb'] if 'rpb' in options else 10
        region_tag = options['region_tag'] if 'region_tag' in options else self.region_tag
        current_tag = options['current_tag'] if 'current_tag' in options else self.current_tag
        current_class = options['current_class'] if 'current_class' in options else self.current_class
        space = options['space'] if 'space' in options else self.space
        space_block = options['space_block'] if 'space_block' in options else space
        prev_btn = options['prev'] if 'prev' in options else self.prev
        prev_class = options['prev_class'] if 'prev_class' in options else self.prev_class
        next_btn = options['next'] if 'next' in options else self.next
        next_class = options['next_class'] if 'next_class' in options else self.next_class
        prev_block = options['prev_block'] if 'prev_block' in options else self.prev_block
        prev_block_class = options['prev_block_class'] if 'prev_block_class' in options else self.prev_block_class
        next_block = options['next_block'] if 'next_block' in options else self.next_block
        next_block_class = options['next_block_class'] if 'next_block_class' in options else self.next_block_class
        first = options['first'] if 'first' in options else self.first
        first_class = options['first_class'] if 'first_class' in options else self.first_class
        last = options['last'] if 'last' in options else self.last
        last_class = options['last_class'] if 'last_class' in options else self.last_class
        link_prefix = options['link_prefix'] if 'link_prefix' in options else self.link_prefix
        link_suffix = options['link_suffix'] if 'link_suffix' in options else self.link_suffix
        link_function = options['link_function'] if 'link_function' in options else self.link_function
        link_params = options['link_params'] if 'link_params' in options else {}

        if link_function:
            link_prefix = 'javascript:%s(\'' % link_function
            link_suffix = '\');'

        uri = self.helper.url.parse(self.request)
        url = uri.path
        params = uri.query
        params = dict(params, **link_params)

        last_page = int(self.helper.math.ceil(total_count * 1.0 / rpp * 1.0))

        if page < 1:
            page = 1
        elif page > last_page:
            page = last_page

        current_block = int(self.helper.math.ceil(page * 1.0 / rpb * 1.0))
        last_block = int(self.helper.math.ceil(last_page * 1.0 / rpb * 1.0))

        current_block = int(current_block)
        last_block = int(last_block)

        output = ''

        if region_tag:
            output = '%s<%s class="%s">' % (output, self.region_tag, self.region_class)

        # First
        if first and current_block > 1:
            params['page'] = 1
            s = self.helper.url.build(url, params)
            output = ('%s<%s href="%s%s%s" class="%s">%s</%s>%s'
                      % (output,
                         self.link_tag,
                         link_prefix,
                         s,
                         link_suffix,
                         first_class,
                         first,
                         self.link_tag,
                         space_block))

        # Prev Block
        if prev_block and current_block > 1:
            params['page'] = ((current_block - 2) * rpb) + 1
            s = self.helper.url.build(url, params)
            output = ('%s<%s href="%s%s%s" class="%s">%s</%s>%s'
                      % (output,
                         self.link_tag,
                         link_prefix,
                         s,
                         link_suffix,
                         prev_block_class,
                         prev_block,
                         self.link_tag,
                         space_block))

        # Prev Button
        if prev_btn and current_block > 1:
            params['page'] = page - 1
            s = self.helper.url.build(url, params)
            output = ('%s<%s href="%s%s%s" class="%s">%s</%s>%s'
                      % (output,
                         self.link_tag,
                         link_prefix,
                         s,
                         link_suffix,
                         prev_class,
                         prev_btn,
                         self.link_tag,
                         space))

        # Paging
        ra = ((current_block - 1) * rpb) + 1
        rb = (current_block * rpb) + 1

        for i in range(ra, rb):
            if i < 1:
                continue
            elif i > last_page:
                break

            if ra == i:
                _space = ''
            else:
                _space = space

            output = '%s%s' % (output, _space)

            if page == i:
                output = ('%s<%s class="%s">%s</%s>'
                          % (output, current_tag, current_class, i, current_tag))
            else:
                params['page'] = i
                s = self.helper.url.build(url, params)

                output = ('%s<%s%s href="%s%s%s">%s</%s>'
                          % (output,
                             self.link_tag,
                             self.link_class,
                             link_prefix,
                             s,
                             link_suffix,
                             i,
                             self.link_tag))

        # Next Button
        if next_btn and current_block < last_block:
            params['page'] = page + 1
            params['page'] = last_page if params['page'] > last_page else params['page']
            s = self.helper.url.build(url, params)

            output = ('%s%s<%s href="%s%s%s" class="%s">%s</%s>'
                      % (output,
                         space,
                         self.link_tag,
                         link_prefix,
                         s,
                         link_suffix,
                         next_class,
                         next_btn,
                         self.link_tag))

        # Next Block
        if next_block and current_block < last_block:
            params['page'] = (current_block * rpb) + 1
            params['page'] = last_page if params['page'] > last_page else params['page']
            s = self.helper.url.build(url, params)

            output = ('%s%s<%s href="%s%s%s" class="%s">%s</%s>'
                      % (output,
                         space_block,
                         self.link_tag,
                         link_prefix,
                         s,
                         link_suffix,
                         next_block_class,
                         next_block,
                         self.link_tag))

        # Last
        if last and last_block > 1 and current_block < last_block:
            params['page'] = last_page
            s = self.helper.url.build(url, params)
            output = ('%s<%s href="%s%s%s" class="%s">%s</%s>%s'
                      % (output,
                         self.link_tag,
                         link_prefix,
                         s,
                         link_suffix,
                         last_class,
                         last,
                         self.link_tag,
                         space_block))

        if region_tag:
            output = '%s</%s">' % (output, self.region_tag)

        return output



# -*- coding: utf-8 -*-


try:
    import cPickle as cPickle
except ImportError:
    import pickle as cPickle

try:
    unicode = unicode
except NameError:
    unicode = str

try:
    long = long
except NameError:
    long = int

from ..cache import CacheDriver as dpCacheDriver
from ..model import ModelSingleton as dpModelSingleton
from ..engine import Engine as dpEngine
from sqlalchemy.exc import OperationalError


class _Empty(object):
    pass


class _Pickled(object):
    pass


pickled = _Pickled()
empty = _Empty()


class SqliteCacheDriver(dpEngine, dpCacheDriver):
    @staticmethod
    def getpool(config_dsn=None, host=None, port=None, database=None, user=None, password=None):
        driver = SqliteCacheDriver(config_dsn=config_dsn)
        driver._config_dsn = config_dsn
        driver._reference_count = 0

        return driver

    def _table_name(self, config_dsn):
        return ('table_%s_%s' % (config_dsn.driver, config_dsn.database)).replace('/', '_').replace('.', '_')

    def create_table(self, config_dsn, clear=False):
        table_name = self._table_name(config_dsn)

        proxy = dpModelSingleton().begin(config_dsn, cache=True)
        proxy.execute("""
            CREATE TABLE IF NOT EXISTS %(table_name)s (
                key TEXT PRIMARY KEY ASC NOT NULL,
                val TEXT,
                type INT,
                expire_at integer DEFAULT NULL
            )
        """ % {'table_name': table_name})
        proxy.execute('CREATE INDEX IF NOT EXISTS %(table_name)s_key_idx ON %(table_name)s (expire_at);'
                      % {'table_name': table_name})

        if clear:
            proxy.execute('DELETE FROM %(table_name)s' % {'table_name': table_name})

        proxy.commit()

    @staticmethod
    def _types():
        return {
            1: str,
            2: unicode,
            3: int,
            4: long,
            5: float,
            6: list,
            7: tuple,
            8: dict,
            9: pickled
        }

    @staticmethod
    def _types_required_serialize():
        return {
            6: list,
            7: tuple,
            8: dict
        }

    @staticmethod
    def _type_to_key(t):
        for key, val in SqliteCacheDriver._types().items():
            if t == val:
                return key

        return None

    @staticmethod
    def _key_to_type(key):
        types = SqliteCacheDriver._types()

        return types[key] if key in types else None

    @staticmethod
    def _val_to_type(val):
        return SqliteCacheDriver._type_to_key(val.__class__)

    def getconn(self):
        return self

    def _referenced(self):
        self._reference_count += 1
        self._try_clear_expired()

    def _try_clear_expired(self):
        if self._reference_count > 100000:
            self._exec_clear_expired()

    def _exec_clear_expired(self):
        self._reference_count = 0

        return dpModelSingleton().execute(
            'DELETE FROM {table_name} WHERE expire_at < ?'.replace('{table_name}', self._table_name(self._config_dsn)),
            self.helper.datetime.current_time(), self._config_dsn, cache=True)

    def get(self, key, retry_count=0, raise_error=False):
        self._referenced()

        try:
            result = dpModelSingleton().row(
                'SELECT '
                '   * '
                'FROM {table_name} WHERE key = ?'.replace('{table_name}', self._table_name(self._config_dsn)),
                key, self._config_dsn, cache=True)
        except OperationalError as e:
            # Retry (hack for sqlite database locking)
            if retry_count:
                import time
                time.sleep(0.02)

                try:
                    self.create_table(self._config_dsn)
                except:
                    pass

                return self.get(key, retry_count - 1)

            else:
                if raise_error:
                    raise e

                else:
                    return None

        if result:
            if result['expire_at'] and result['expire_at'] < self.helper.datetime.current_time():
                return None
            else:
                type = self._key_to_type(result['type'])

                if result['type'] in self._types_required_serialize():
                    try:
                        ret = self.helper.json.deserialize(result['val'], raise_exception=True)

                        if type is tuple:
                            return tuple(ret)

                        else:
                            return ret

                    except:  # Failed to json deserialization
                        self.logger.exception(result)
                        return None

                elif type is pickled:
                    try:
                        if self.helper.system.py_version <= 2:
                            return cPickle.loads(self.helper.string.to_str(result['val']))
                        else:
                            return cPickle.loads(result['val'])

                    except:  # Failed to unpickling
                        self.logger.exception(result)
                        return None

                elif type is str:
                    return self.helper.string.to_str(result['val'])

                elif type is unicode:
                    return self.helper.string.to_unicode(result['val'])

                elif type is int:
                    return int(result['val'])

                elif type is long:
                    return long(result['val'])

                elif type is float:
                    return float(result['val'])

                else:
                    return result['val']

        else:
            return None

    def set(self, key, val, expire_in=1209600, retry_count=10, raise_error=False):
        if expire_in is not None:
            expire_in = self.helper.datetime.current_time() + expire_in

        type = self._val_to_type(val)
        val_serialized = empty

        if not type or type in self._types_required_serialize():
            try:
                try:
                    if type:
                        val_serialized = self.helper.json.serialize(val, raise_exception=True)

                    else:
                        raise Exception()

                except:
                    val_serialized = cPickle.dumps(val)
                    type = self._type_to_key(pickled)

            except:
                pass

        else:
            val_serialized = str(val)

        if val_serialized == empty:
            return False

        try:
            if expire_in:
                ret = dpModelSingleton().execute(
                    'INSERT OR REPLACE INTO {table_name} (key, val, type, expire_at) '
                    '   VALUES (?, ?, ?, ?)'
                    .replace('{table_name}', self._table_name(self._config_dsn)),
                    (key, val_serialized, type, expire_in), self._config_dsn, cache=True)

            else:
                ret = dpModelSingleton().execute(
                    'INSERT OR REPLACE INTO {table_name} (key, val, type, expire_at) '
                    '   VALUES (?, ?, ?, (SELECT expire_at FROM {table_name} WHERE key = ?))'
                    .replace('{table_name}', self._table_name(self._config_dsn)),
                    (key, val_serialized, type, key), self._config_dsn, cache=True)

            return ret

        except OperationalError as e:
            # Retry (hack for sqlite database locking)
            if retry_count:
                import time
                time.sleep(0.02)

                try:
                    self.create_table(self._config_dsn)
                except:
                    pass

                return self.set(key, val, expire_in, retry_count - 1)

            else:
                if raise_error:
                    raise e

                else:
                    return False

    def delete(self, key):
        return dpModelSingleton().execute("""
            DELETE FROM {table_name}
                WHERE
                    key = ?""".replace('{table_name}', self._table_name(self._config_dsn)),
            key, self._config_dsn, cache=True)

    def increase(self, key, amount, expire_in=None):
        if expire_in is not None:
            expire_in = self.helper.datetime.current_time() + expire_in

        if expire_in:
            return dpModelSingleton().execute(
                'UPDATE {table_name} '
                '   SET '
                '       val = val + ?,'
                '       expire_at = ? '
                '   WHERE '
                '       key = ?'.replace('{table_name}', self._table_name(self._config_dsn)),
                (amount, expire_in, key), self._config_dsn, cache=True)

        else:
            return dpModelSingleton().execute(
                'UPDATE {table_name} '
                '   SET '
                '       val = val + ?'
                '   WHERE '
                '       key = ?'.replace('{table_name}', self._table_name(self._config_dsn)),
                (amount, key), self._config_dsn, cache=True)

    def ttl(self, key):
        record = dpModelSingleton().row("""
                SELECT
                    *
                FROM
                    {table_name}
                WHERE
                    key = ?""".replace('{table_name}',
                                       self._table_name(self._config_dsn)), key, self._config_dsn, cache=True)

        if not record:
            return -2
        elif not record['expire_at']:
            return -1
        else:
            return record['expire_at'] - self.helper.datetime.time()
